/*
 ===========================================================================
 =                                                                         =
 =  (C) Copyright 1991-1994 The Trustees of Indiana University             =
 =                                                                         =
 =  Permission to use, copy, modify, and distribute this program for       =
 =  non-commercial use and without fee is hereby granted, provided that    =
 =  this copyright and permission notice appear on all copies and          =
 =  supporting documentation, the name of Indiana University not be used   =
 =  in advertising or publicity pertaining to distribution of the program  =
 =  without specific prior permission, and notice be given in supporting   =
 =  documentation that copying and distribution is by permission of        =
 =  Indiana University.                                                    =
 =                                                                         =
 =  Indiana University makes no representations about the suitability of   =
 =  this software for any purpose. It is provided "as is" without express  =
 =  or implied warranty.                                                   =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 = File:                                                                   =
 =   IUPOP3_UTILITY.C                                                      =
 =                                                                         =
 = Synopsis:                                                               =
 =   This file contains miscellaneous functions that support the           =
 =   VMS IUPOP3 server.                                                    =
 =                                                                         =
 = Authors:                                                                =
 =   Jacob Levanon & Larry Hughes                                          =
 =   Indiana University                                                    =
 =   University Computing Services, Network Applications                   =
 =                                                                         =
 = Credits:                                                                =
 =   This software is based on the Post Office Protocol version 3,         =
 =   as implemented by the University of California at Berkeley.           =
 =                                                                         =
 ===========================================================================
 =                                                                         =
 =  This file based on version 1.8-1 of the iupop3 server, the             =
 =  last version made made Indiana University.                             =
 =                                                                         =
 =  For differences to this version read changes.txt                       =
 =                                                                         =
 = M. Stenns         Jan 1997                                              =
 =                                                                         =
 ===========================================================================
*/

/* ====================================================================== */
/* Includes */
/* ====================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>             /* for va_ stuff */

#include <unixio.h>
#ifndef MULTINET
#include <iodef.h>
#endif
#ifdef UCX
#include <in.h>
#else
#include <sys/types.h>
#include <netinet/in.h>
#endif

#ifdef KERBEROS
#include <krb.h>
#include <des.h>
#endif /* KERBEROS */

#include "iupop3_general.h"
#include "iupop3_vms.h"
#include "iupop3_global.h"
#include "version.h"
#include "md5.h"

/* ====================================================================== */
/* Prototypes */
/* ====================================================================== */


/* ====================================================================== */
/* Get Date */
/* ====================================================================== */
void get_date (char *date_str)
{
  time_t calendar_time;
  struct tm *local_time;

  calendar_time = time(0);
  local_time = localtime(&calendar_time);
  ++local_time->tm_mon;
  sprintf(date_str,"19%02d-%02d-%02d",
          local_time->tm_year,local_time->tm_mon,local_time->tm_mday);
}

/* ====================================================================== */
/* Get Time */
/* ====================================================================== */
void get_time(char *time_str)
{
  time_t calendar_time;
  struct tm *local_time;

  calendar_time = time(0);
  local_time = localtime(&calendar_time);
  sprintf(time_str,"%02d:%02d:%02d",
          local_time->tm_hour,local_time->tm_min,local_time->tm_sec);
}

/* ======================================================================= */
/* Make Argv */
/* ======================================================================= */
void make_argv(char *string, int *argc, char *argv[], char *delimiters)
{
#define isdelimiter(a,b) (strchr(b,a) != NULL)
  char *cpointer;
  char **argpointer = argv;

  *argc = 0;
  cpointer = string;
  while (*cpointer)
  {
    while (*cpointer && isdelimiter(*cpointer,delimiters)) cpointer++;
    if (*cpointer == '\0') break;
    if (*cpointer == '\"')
    {
      cpointer++;
      *argpointer++ = cpointer;
      (*argc)++;
      while (*cpointer && (*cpointer != '\"')) cpointer++;
    }
    else
    {
      *argpointer++ = cpointer;
      (*argc)++;
      while (*cpointer && !isdelimiter(*cpointer,delimiters)) cpointer++;
    }
    if (*cpointer == '\0') break;
    *(cpointer++) = '\0';
  }
  *(argpointer++) = 0;
}


/* ====================================================================== */
/* POP Build_info */
/* ====================================================================== */
int pop_build_info (POP *p)
{
    char buffer[BUFSIZ];
    Message *mp;
    int msg_num;
    int status;

    /*
    **  Initialize mail folder status variables.
    */
    p->msgs_deleted = 0;
    p->last_msg = 0;
    p->bytes_deleted = 0;

    /*
    **  Allocate memory for message information structures.
    */
    p->mptr = calloc ((unsigned)p->msg_count,sizeof(Message));
    if (p->mptr == NULL)
    {
        p->msg_count = 0;
        return (pop_msg (p,POP_FAILURE,
            "Can't build message list for '%s': Out of memory", p->user));
    }

    /*
    **  Scan the NEMAIL folder in the VMS MAIL file, loading the message
    **  information list with information about each message.
    */
    mp = p->mptr;
    for (msg_num=1; msg_num <= p->msg_count; msg_num++, mp++)
    {
        mp->del_flag = FALSE;
        mp->retr_flag = FALSE;
        mp->foreign_msg = FALSE;
        mp->number = msg_num;
        status = mail_message_info (p, &msg_num);
        if (vms_error(status))
          return(pop_msg(p, POP_FAILURE, "%s", vms_message(status)));
	mp->lines++;
        p->newmail_size += mp->length;
    }

    return POP_SUCCESS;
}

/* ====================================================================== */
/* POP Log */
/* ====================================================================== */
void pop_log (int log_level, POP * p, const char *format, ...)
{
    va_list       ap;
    char	  *mp;
    char          message[MAXLINELEN];
    char          date_str[11];
    char          time_str[9];

    va_start(ap, format);

    if (log_level <= master_log_level)
    {
      get_time(time_str);
      get_date(date_str);
      sprintf(message, "%s %s ", date_str, time_str);
      mp = message + strlen(message);
      sprintf(mp, "thread %2d: ", p->threadnum);
      mp = message + strlen(message);
      vsprintf(mp, format, ap);
      fprintf(log_file, "%s\n", message);
      issue_logfile_flush_timer ();
    }

    va_end(ap);
}

/* ====================================================================== */
/* POP Message */
/* ====================================================================== */
int pop_msg(POP *p, int stat, const char *format, ...)
{
    va_list         ap;
    char	    *mp;
    char            message[MAXLINELEN];

    va_start(ap, format);

    mp = message;
    if (stat == POP_SUCCESS)
      sprintf (mp,"%s ",POP_OK);
    else
      sprintf (mp,"%s ",POP_ERR);
    mp += strlen(mp);
    if (format) vsprintf(mp, format, ap);
    va_end(ap);

    netprintf(p, "%s\r\n", message);
    pop_log(LOG_DEBUG, p, "tx: %s", message);
    return(stat);
}

/* ====================================================================== */
/* Patch From Line */
/* ====================================================================== */
void patch_from_line (POP *p, char *from_line)
{
  int  num_tokens;
  char input[256];
  char output[256];
  char *tokens[128];  /* make it huge or the stack might get corrupted */
  char *pointer;
  int all_lower = TRUE;

#if 0
  /* check if from_line has upper chars */
  for (pointer=from_line; *pointer; pointer++)
  {
    if (isupper(*pointer)) 
    {
      all_lower = FALSE; 
      break;
    }
  }
#endif

  /* Token #1 is address, #2 (if it exists) is personal name */
  strcpy (input, from_line);
  make_argv(input, &num_tokens, tokens, " ");

  if ((num_tokens < 1) || (num_tokens > 2))
  {
    pop_log(LOG_ERROR, p, "error parsing %s", from_line);
    return;
  }

  /* Lower case the address (but not the personal name) */
  lower (tokens[0]);

  /*
   * If address is of the form:
   *
   *    FOO%"someone@somewhere"    -or-
   *    NODE::"someone@somewhere"
   *
   * Just use what is in the quotes.
   */
  if (pointer = strchr(tokens[0], '"'))
  {
    p->decnet = FALSE;
    strncpy (output, pointer+1, sizeof output);
    pointer = output + strlen(output) - 1;
    if (*pointer == '"') *pointer = '\0';
  }

  /*
   * If address is of the form:
   *
   *   NAMESPACE:.NODE::USER    (DECnet/OSI format)
   *   NODE::USER               (DECnet phase IV format)
   *
   * Convert it to the form user@node
   */
  else if (pointer = strstr(tokens[0], "::"))
  {
    char *str1 ,*str2, logname[256];

    p->decnet = TRUE;
                                         /* Skip over Phase V namespace */
    if ((str1 = strstr(tokens[0], ":.")) != NULL)
       str1 = str1 + 2;
    else
       str1 = tokens[0];
    strncpy (output, pointer+2, sizeof output);
    strcat (output, "@");
    *pointer = '\0';
                                          /* use logicals to map nodename */
    strncpy (logname,"IUPOP3_NODE_", sizeof logname);
    strcat (logname, str1);
    upper (logname);
    if ((str2 = getenv (logname)) != NULL)
       strcat (output, str2);
    else
       strcat (output, str1);
  }

  /*
   * None of the above; it's local, so append "@hostname"
   * This could both internet (smtp) or decnet messages
   * The default type is set with the -default_type commandline option
   */
  else
  {
    sprintf(output, "%s@%s", tokens[0], myhostname);
  }

  /*
   * Put the address in angle brackets, and prepend personal name if it exists.
   */
  from_line[0] = '\0';
#ifdef PERSONAL_NAME
  if (tokens[1]) sprintf(from_line, "\"%s\" ", tokens[1]);
#endif /* PERSONAL_NAME */
  if (*output == '<')
    sprintf(from_line+strlen(from_line), "%s", output);
  else
    sprintf(from_line+strlen(from_line), "<%s>", output);
#ifndef IGNORE_MAIL11_HEADERS
  p->decnet = TRUE;
#endif
  pop_log(LOG_DEBUG, p, "this is a %s message",p->decnet ? "DECNET":"INTERNET");
}

/* ======================================================================== */
/*  Perform a case-insensitive string comparision */
/* ======================================================================== */
int strncasecmp (char *str1, char *str2, int len)
{
    int i;
    char a, b;

    for (i=len-1;i>=0;i--)
    {
        a = str1[i];
        b = str2[i];
        if (isupper(a)) a = tolower(str1[i]);
        if (isupper(b)) b = tolower(str2[i]);
        if (a > b) return (1);
        if (a < b) return(-1);
    }
    return(0);
}

/* ======================================================================== */
/* System Log */
/* ======================================================================== */
void system_log (int log_level, char *format, ...)
{
    va_list       ap;
    char          *mp;
    char          message[MAXLINELEN];
    char          date_str[11];
    char          time_str[9];

    va_start(ap, format);

    if (log_level <= master_log_level)
    {
      get_time(time_str);
      get_date(date_str);
      sprintf(message, "%s %s ", date_str, time_str);
      mp = message + strlen(message);
      vsprintf(mp, format, ap);
      fprintf(log_file, "%s\n", message);
      issue_logfile_flush_timer ();
    }

    va_end(ap);
}

/* ======================================================================== */
/* Net Printf */
/* ======================================================================== */
int netprintf (POP *p, char *format, ...)
{
  va_list       ap;
  char          message[MAXLINELEN];
  int           bytes = 0;

  va_start(ap, format);

  vsprintf(message, format, ap);
  va_end(ap);

  if (!p->retrieve.blocked)
  {
    bytes = netwrite(p->sockfd, message, strlen(message));
    if (bytes > 0)
    {
      net_bytes_written += bytes;
      net_write_count++;
    }
  }
  return bytes;
}


/* ====================================================================== */
/*  writes the data in p->retrieve.send_buffer to the net */
/* ====================================================================== */
int netwrite_send_buffer (POP *p)
{
  int bytes = 0;

  if (p->retrieve.send_bufsize && !p->retrieve.blocked)
  {
    bytes = netwrite (p->sockfd, p->retrieve.send_buffer, p->retrieve.send_bufsize);
    if (bytes > 0)
    {
      net_bytes_written += bytes;
      net_write_count++;
    }
  }
  p->retrieve.send_bufsize = 0;
  return bytes;
}

/* ======================================================================== */
/* copy text from p->retrieve.buffer to p->retrieve.send_buffer */
/* ======================================================================== */
int sndbuf_cpy (POP *p)
{
int length = 0;
char * ptr = p->retrieve.send_buffer + p->retrieve.send_bufsize;

  /* reserve place for multiline termination and "byte stuff" */
  if (((SEND_BUFFERSIZE - 8) - p->retrieve.send_bufsize) > p->retrieve.bufsize)
  {
    p->retrieve.buffer[p->retrieve.bufsize++] = '\r';
    p->retrieve.buffer[p->retrieve.bufsize++] = '\n';
    p->retrieve.buffer[p->retrieve.bufsize] = '\0';
    /* "Byte stuff" if necessary */
    if (p->retrieve.buffer[0] == POP_TERMINATE)
    {
      *ptr++ = POP_TERMINATE;
      p->retrieve.send_bufsize++;
    }
    strcpy (ptr, p->retrieve.buffer);
    p->retrieve.send_bufsize += p->retrieve.bufsize;
    length = p->retrieve.bufsize;
    p->retrieve.bufsize = 0;
  }
  return length;
}

/* ====================================================================== */
/*  allocate  send buffer */
/* ====================================================================== */
int allocate_send_buffer (POP *p)
{
  int status = TRUE;

  if (!p->retrieve.send_buffer) 
  {
    p->retrieve.send_buffer = calloc (1, SEND_BUFFERSIZE * sizeof (char));
    if (!p->retrieve.send_buffer) status = FALSE;
  }
  return status;
}

/* ======================================================================== */
/* lower */
/* ======================================================================== */
char * lower (char * str)
{
char * tmp;

 /* convert to lower */
 if (str)
 {
    for (tmp = str; *tmp; tmp++) *tmp = tolower (*tmp);
 }
 return str;
}

/* ======================================================================== */
/* upper */
/* ======================================================================== */
char * upper (char * str)
{
char * tmp;

 /* convert to upper */
 if (str)
 {
    for (tmp = str; *tmp; tmp++) *tmp = toupper (*tmp);
 }
 return str;
}

/* ======================================================================== */
/*  Replaces tabs through spaces */
/* ======================================================================== */
char * replace_tabs (char * string)
{
  char * pointer = string;

  if (string) 
  {
    while ((pointer = strchr(pointer,'\t'))) *pointer = ' ';
  }
  return string;
}

/* ======================================================================== */
/* checks the digest sent by the client against timestamp and password      */
/* Return:	 0 (fail) or 1 (ok)                                         */
/* ======================================================================== */
int check_digest (unsigned int start_time[2], char *secret, char *digest_client)
{
  char digest_hexstring [34];
  MD5_CTX context;
  unsigned char digest[16];
  unsigned char timestamp[80];
  char *tmp;
  int i;
  int status = 0;  /* 0 means wrong password */

  sprintf ((char *) timestamp, TIMESTAMP_FORMAT,
           VERSION, myhostname, time_from_bintime (start_time, 0));
  if (timestamp && secret && digest_client && (strlen(digest_client) >= 32))
  {
    MD5Init (&context);
    MD5Update (&context, timestamp, strlen ((char *) timestamp));
    MD5Update (&context,(unsigned char *) secret, strlen (secret));
    MD5Final (digest, &context);
    for (i=0; i<16; i++) sprintf (&digest_hexstring[2*i],"%02.2x",digest[i]);
    if (!strncmp (digest_hexstring,digest_client,32)) status = 1;
  }
  return status;
}
